#include "WirelessController.h"

WirelessController::WirelessController( unsigned long * marks_duration,
										int *message_code_start,
										int *message_code_one,
										int *message_code_zero,
										int *message_code_end,
										byte number_of_marks, 
										byte number_of_marks_start, 
										byte number_of_marks_bit, 
										byte number_of_marks_end,
										byte mark_tolerance,
										byte encoding_type,
										byte input_type,
										int number_of_bits)
{
	this->marks_duration = marks_duration;
	this->message_code_start = message_code_start;
	this->message_code_one = message_code_one;
	this->message_code_zero = message_code_zero;
	this->message_code_end = message_code_end;
	this->number_of_marks = number_of_marks;
	this->number_of_marks_start = number_of_marks_start;
	this->number_of_marks_bit = number_of_marks_bit;
	this->number_of_marks_end = number_of_marks_end;
	this->mark_tolerance = mark_tolerance;
	this->encoding_type = encoding_type;
	this->input_type = input_type;
	this->number_of_bits = number_of_bits;
}

void WirelessController::intializeReading(byte read_pin)
{
	pinMode(read_pin, INPUT);
	this->read_pin = read_pin;
	this->last_input_value = 0;
	this->swap_time = 0;
	this->last_swap_time = 0;
	this->reception_status = RECEIVING_START;
	this->reception_position = 0;
	code_buff = (byte*)malloc(((number_of_bits+7)/8)*sizeof(byte));
	message_bit_marks_buffer = (int*)malloc((number_of_marks_bit)*sizeof(int));
}

void WirelessController::intializeWritting(byte write_pin, boolean modulate_output)
{
	pinMode(write_pin, OUTPUT);
	this->write_pin = write_pin;
	this->modulate_output = modulate_output;
}

void WirelessController::sendCode(byte *message_buffer)
{
    end_last_code = micros();
	// Sends the header of the message
	for(int i=0; i<number_of_marks_start; i++) 
		sendMark(message_code_start[i]);
	// Sends all of the bits in the message
	sendBits(message_buffer, number_of_bits);
	// And finally sends the ending of the message
	for(int i=0; i<number_of_marks_end; i++) 
		sendMark(message_code_end[i]);
}

// Send a specified number of bits from the bit buffer
void WirelessController::sendBits(byte *bit_buffer, int count)
{
   for (int i=0; i<count; i++)
    {
        if (bit_buffer[i/8] & (0x80 >> (i%8)) ) 
        {
			for (int j=0; j<number_of_marks_bit; j++)
			{
				sendMark(message_code_one[j]);
			}
        }
        else
        {
			for (int j=0; j<number_of_marks_bit; j++)
			{
				sendMark(message_code_zero[j]);
			}
        }
    }
}

// Functions to send data 
void WirelessController::sendMark(int code)
{
	// Calculates the time to stop transmitting this mark
	// If the code has a negative value means the mark has a LOW value
	boolean mark_low = false;
	unsigned long end_time_mark;
	if (code<0) 
	{
		mark_low = true;
		end_time_mark =  micros()+ (unsigned long) marks_duration[(-code)-1]; 
	}
	else 
		end_time_mark =  micros()+ (unsigned long) marks_duration[code-1]; 

	// If we're transmitting a LOW mark we just need to wait, doesn't matter if the
	// signal is modulated or not. SO we just set the pin LOW and wait
	if (mark_low)
	{		
		digitalWrite(write_pin, LOW);	
		while (micros()<end_time_mark);	
	}
	// If we're sending a HIGH value, we need to check is the signal is modulated
	else
	{
		// If it's not, we just set the pin HIGH and wait
		if (!modulate_output)
		{
			digitalWrite(write_pin, HIGH);	
			while (micros()<end_time_mark);	
		}
		// If the output is modulated, let's make it work around the 36Khz range
		else
		{
			while (micros() < end_time_mark)
			{
				digitalWrite(write_pin, HIGH);	
				delayMicroseconds(10);    
				digitalWrite(write_pin, LOW);	
				delayMicroseconds(10);	
			}	
		}
	}
	end_last_code = end_time_mark;
}

// Compares the given coce with the last code received
boolean WirelessController::compareCode(byte* code)
{
    for (int i=0; i<((number_of_bits+7)/8); i++)
	{
		if (code[i] != code_buff[i]) return false;
	}
    return true;
}

// Copies the last received code to an external byte buffer
void WirelessController::getCode(byte* code)
{
	memcpy(code, code_buff, (number_of_bits+7)/8);
}

int WirelessController::readMark()
{
	// Reads the data differently depending on the input type
	int current_value;
	if (input_type == INPUT_ACTIVE_HIGH)
		current_value = digitalRead(read_pin);
	else
		current_value = !digitalRead(read_pin);
	// It's only relevanto to do anything when there is a change in the signal	
    if (current_value == last_input_value)
        return MARK_NO_MARK;
    // Saves the time the signal has changed
    last_swap_time = swap_time;
    swap_time = micros();
    last_input_value = current_value;
    unsigned long signal_time = swap_time - last_swap_time;
	// Checks if this mark matches any of the known mark lenghts
	byte mark_number;
	unsigned long current_mark_tolerance;
	for (mark_number = 0; mark_number < number_of_marks; mark_number++)
	{
		// We compare the timings with values within a percentual error tolerance
		current_mark_tolerance = (unsigned long)marks_duration[mark_number]*(unsigned long)mark_tolerance/100.0;
		// If a mark was found, checks is it's a high or low mark and returns it's number
		if (signal_time>marks_duration[mark_number]-current_mark_tolerance && 
			signal_time<marks_duration[mark_number]+current_mark_tolerance)
		{			
			if (current_value == 0)
				return mark_number+1;
			else
				return -(mark_number+1);
		}
    }
	// Ir no matching marks were found, then there was as error
    return MARK_ERROR;            
}

boolean WirelessController::processIncomingBit(byte next_bit)
{
	reception_position = 0;		
	byte_value = byte_value | (next_bit << (7-bit_number%8));
	bit_number++; 
	// If this bit is the end of a byte or the last bit in the message, saves it's value
	if ( (bit_number % 8 == 0 ) || ( bit_number == number_of_bits) )
	{
		code_buff[((bit_number+7)/8)-1] = byte_value;
		byte_value = 0;
	}
	// If this bit was the last one, checks if there are any marks to 
	// be received to end the message
	if (bit_number == number_of_bits )
	{
		// If there is only mark in the end of the message, it should be the break
		// that separates messages... So we will assume the message has ended and get
		// ready to start receiving the next one
		if (number_of_marks_end <= 1)
		{
			reception_status = RECEIVING_START;
			return true;
		}
		else
		{
 			reception_status = RECEIVING_END;
		}
	}
	return false;
}

boolean WirelessController::codeReceivedPulseWidth()
{ 
    int current_mark = readMark();
	// Keeps checking if there are any new incoming marks
    if (current_mark != MARK_NO_MARK) 
    {	
		if (reception_status == RECEIVING_START)
		{
			// Checks if the right marks for the header are being received
			if (message_code_start[reception_position++] == current_mark)
			{
				// When it receives the last mark, it's time to start receiving bits
				if (reception_position == number_of_marks_start)
				{
					reception_status = RECEIVING_BITS;
					reception_position = 0;
					bit_number = 0; 
					byte_value = 0; 
				} 
			}
			// In case of error, starts waiting for the beginning of the header again
			else
			{
				reception_position = 0;
				return false;
			}
		}
		else if (reception_status == RECEIVING_BITS)
		{
			message_bit_marks_buffer[reception_position++] = current_mark;
			// If we receives enough marks for a full bit, checks if it's a 0 or a 1
			if (reception_position == number_of_marks_bit)
			{
				byte i;
				// First checks if the marks match the marks for a 1
				for (i=0; i<number_of_marks_bit; i++)			
				{
					if (message_code_one[i] != message_bit_marks_buffer[i])
						break;
				}
				// If the break didn't happen, means the bit received was a 1
				if (i==number_of_marks_bit)
				{
					return processIncomingBit(1);
				}
				// If it wasn't a 1, checks if it's a 0...
				for (i=0; i<number_of_marks_bit; i++)
				{
					if (message_code_zero[i] != message_bit_marks_buffer[i])
						break;
				}
				// If the break didn't happen, means the bit received was a 0
				if (i==number_of_marks_bit)
				{
					return processIncomingBit(0);
				}
				// If the code reached this point, means that the marks didn't match
				// a 0 or a 1, so there was an error
				reception_status = RECEIVING_START;
				reception_position = 0;
				return false;
			}		
		}
		else if (reception_status == RECEIVING_END)
		{
			// Checks if the right marks for the ending of the message are being received
			if (message_code_end[reception_position++] == current_mark)
			{
				// When it receives the last mark, it's time to start receiving bits
				if (reception_position == number_of_marks_end)
				{
					reception_status = RECEIVING_START;
					reception_position = 0;
					return true;
				}
			}
			// In case of error, starts waiting for the beginning of the header again
			else
			{
				reception_status = RECEIVING_START;
				reception_position = 0;
				return false;
			}
		}
    }    
    return false;
}

boolean WirelessController::codeReceivedManchester()
{
    int current_mark = readMark();
	// Keeps checking if there are any new incoming marks
    if (current_mark != MARK_NO_MARK) 
    {	
		// If we're at the beginning of the message, let's check if there is any 
		// header to be read. If there is no header, skips to reading the bits
		if (reception_status == RECEIVING_START && number_of_marks_start <1)
		{
			reception_status = RECEIVING_BITS;
			reception_position = 0;
			bit_number = 0; 
			byte_value = 0; 
			processIncomingBit(1);
			bit_value = 1;
		}
		if (reception_status == RECEIVING_START)
		{
			// Checks if the right marks for the header are being received
			if (message_code_start[reception_position++] == current_mark)
			{
				// When it receives the last mark, it's time to start receiving bits
				if (reception_position == number_of_marks_start-1)
				{
					reception_status = RECEIVING_BITS;
					reception_position = 0;
					bit_number = 0; 
					byte_value = 0;
					bit_value = 1;
				}
			}
			// In case of error, starts waiting for the beginning of the header again
			else
			{
				reception_position = 0;
				return false;
			}
		} 
		else if (reception_status == RECEIVING_BITS)
		{
			// Makes sure all the marks haave positive values, since now we don't care 
			// if the marks have HIGH ou LOW values, we're just looking at the duration
			int aux_code = message_code_one[0];
			if (message_code_one[0]<0) aux_code = -aux_code;
			if (current_mark<0) current_mark = -current_mark;
			// If the mark is short it means the nexy bit remains the same, 
			// assuming the next mark is another short one
			if (current_mark == aux_code)
			{
				reception_status = RECEIVING_MANCHESTER_END;
				return false;
			}
			// If the mark is long , means the next bit is changing
			else if (current_mark == aux_code+1)
			{
				bit_value = (bit_value+1)%2;
				return processIncomingBit(bit_value);			
			}
			// If the code reached this point, means that the marks didn't match
			// a 0 or a 1, so there was an error, resets the sequence
			reception_status = RECEIVING_START;
			reception_position = 0;
			return false;		
		}
		else if (reception_status == RECEIVING_MANCHESTER_END)
		{	
			int aux_code = message_code_one[0];
			if (aux_code<0) aux_code = -aux_code;
			if (current_mark<0) current_mark = -current_mark;
			if (current_mark == aux_code )
			{
				reception_status = RECEIVING_BITS;
				return processIncomingBit(bit_value);
			}
			// If the code reached this point, means that the mark wasn't
			// another short one, so there was an error somewere
			reception_status = RECEIVING_START;
			reception_position = 0;
			return false;
		}
		else if (reception_status == RECEIVING_END)
		{
			// Checks if the right marks for the ending of the message are being received
			if (message_code_end[reception_position++] == current_mark)
			{
				// When it receives the last mark, it's time to start receiving bits
				if (reception_position == number_of_marks_end)
				{
					reception_status = RECEIVING_START;
					reception_position = 0;
					return true;
				}
			}
			// In case of error, starts waiting for the beginning of the header again
			else
			{
				reception_status = RECEIVING_START;
				reception_position = 0;
				return false;
			}
		}
    }
    return false;
}

boolean WirelessController::codeReceived()
{
	// Tries to read the message usign the right encoding
	if (encoding_type == ENCODING_PULSE_WIDTH) return codeReceivedPulseWidth();
	else if (encoding_type == ENCODING_MANCHESTER) return codeReceivedManchester();
	return false;
}

/* Debug functions, to be removed 

void WirelessController::buffPrint(char next_char)
{
	charbuff[buff_pos++] = next_char;
	if (buff_pos>=BUFFER_SIZE-2)
	{
		charbuff[BUFFER_SIZE-2] = 0;
		charbuff[BUFFER_SIZE-1] = 0;
		Serial.println(charbuff);
		buff_pos = 0;
	}
}

void WirelessController::codeCheckMarks()
{
    int current_mark = readMark();
	// Keeps checking if there are any new incoming marks
    if (current_mark != MARK_NO_MARK) 
    {
		if (current_mark == MARK_ERROR) buffPrint ('-');
		else if (current_mark>0) buffPrint('A'+current_mark-1);
		else if (current_mark<0) buffPrint('a'-current_mark-1);		
	}
}
/**/